{ AI Unit Designed by Jason McMillen  2002-2003 }
unit AIUnit;

interface

uses Math, Dialogs, SysUtils;

type
  TNeuron = class(TObject)
    DigitalOutput: Boolean;
    NumberOfInputs: Integer;
    Weight: Array of Real;{Double}
    constructor Init(NumOfInputs: Integer; Digital: Boolean);
    function Output(Inputs: Array of Real): Real;
  end;
  TNeuronLayer = class(TObject)
    NumberOfNeurons: Integer;
    Neurons: Array of TNeuron;
    constructor Init(NumOfNeurons, NumOfInputsPerNeuron: Integer; Digital: Boolean);
    procedure Execute(Inputs: Array of Real; var Outputs: Array of Real);
  end;
  TNeuralNet = class(TObject)
   private
    NumberOfInputs: Integer;
    NumberOfOutputs: Integer;
    NumberOfLayers: Integer;
    //storage for each layer of neurons including the output layer
    Layers: Array of TNeuronLayer;
   public
    FitnessScore: Integer;
    DigitalOutput: Boolean;
    constructor Init(NumOfInputs: Integer; NeuronSet: Array of Integer; Digital: Boolean);
    destructor Kill;
    function GetNumberOfWeights: Integer;                 //returns the total number of weights in the net
    procedure GetWeights(var Weights: Array of Real);     //gets the weights from the NN
    procedure PutWeights(Weights: Array of Real);         //replaces the weights with new ones
//    function Sigmoid(Activation, Response: Real): Real;   //sigmoid response curve
    procedure Execute(Input: Array of Real; var Output: Array of Real);//calculates the outputs from a set of inputs
  end;
  TChromosome = class(TObject)
    NumberOfValues: Integer;
    Values: Array of Real;{Double}
    FitnessScore: Integer;
    constructor Init(Size: Integer);
    destructor DeInit;
    procedure GetChromo(var Chromo: Array of Real);
    procedure PutChromo(Chromo: Array of Real);
  end;

const
  CrossOver_Rate = 0.7;
  Mutation_Rate = 0.05;{0.2 to 0.05 recommended}
  Digital_Mutation_Rate = 0.001;
  ThresholdCheck_On = True;

function Sigmoid(Activation, Response: Real): Real;

procedure GetChromosomes(NeuralNets: Array of TNeuralNet; var Population: Array of TChromosome; ChromoSize: Integer);
procedure ApplyAlgorithm(var Population: Array of TChromosome; NumberOfPopulation, ChromoSize: Integer);
procedure PutChromosomes(var NeuralNets: Array of TNeuralNet; Population: Array of TChromosome; ChromoSize: Integer);

implementation

function Sigmoid(Activation, Response: Real): Real;
const e = 2.7183;
begin
     Result := 1 / (1 + Power(e, (Activation / Response)));
end;

{-- Genetic Algorithm Functions --}
procedure GetChromosomes(NeuralNets: Array of TNeuralNet; var Population: Array of TChromosome; ChromoSize: Integer);
var i: Integer;
    TempValues: Array of Real;
begin
     if (Length(NeuralNets) > 0) then
        if (Length(NeuralNets) = Length(Population)) then
           for i := 0 to Length(NeuralNets) - 1 do
           begin
                Population[i].FitnessScore := NeuralNets[i].FitnessScore;
                SetLength(TempValues, ChromoSize);
                NeuralNets[i].GetWeights(TempValues);
                Population[i].PutChromo(TempValues);
           end
        else
        begin
             ShowMessage('ERROR <GetChromosomes()>: Non-Equal Array Sizes!');
             Halt;
        end;

     TempValues := nil;
end;
procedure PutChromosomes(var NeuralNets: Array of TNeuralNet; Population: Array of TChromosome; ChromoSize: Integer);
var i: Integer;
begin
     if (Length(NeuralNets) > 0) then
        if (Length(NeuralNets) = Length(Population)) then
           for i := 0 to Length(NeuralNets) - 1 do
           begin
                NeuralNets[i].FitnessScore := Population[i].FitnessScore;
                NeuralNets[i].PutWeights(Population[i].Values);
           end
        else
        begin
             ShowMessage('ERROR <PutChromosomes()>: Non-Equal Array Sizes!');
             Halt;
        end;
end;
procedure ApplyAlgorithm(var Population: Array of TChromosome; NumberOfPopulation, ChromoSize: Integer);
var i, j, k: Integer;
    NewPopulation: Array of TChromosome;
    TotalFitness: Integer;
    RandomSelectionA, RandomSelectionB: Real;
    SelectionA, SelectionB: Array of Real;
    AFitness, BFitness: Integer;
    SplitPos: Integer;
begin
     {- Initalizing NewPopulation and getting Total Fitness -}
     SetLength(NewPopulation, Length(Population));
     TotalFitness := 0;
     for i := 0 to NumberOfPopulation - 1 do
     begin
          NewPopulation[i] := TChromosome.Init(ChromoSize);

          inc(Population[i].FitnessScore);{Every creature has 1 peice of the pie!}

          inc(TotalFitness, Population[i].FitnessScore);
     end;

     SetLength(SelectionA, ChromoSize);//Another Address problem!!!
     SetLength(SelectionB, ChromoSize);
     for i := 0 to NumberOfPopulation - 1 do
     begin
     {- Roulette Wheel Selection -}
          {Select A}
          RandomSelectionA := Random * TotalFitness;
          for j := 0 to NumberOfPopulation - 1 do
          begin
               if (RandomSelectionA < Population[j].FitnessScore) then
               begin
                    for k := 0 to Population[j].NumberOfValues - 1 do
                        SelectionA[k] := Population[j].Values[k];
                    AFitness := Population[j].FitnessScore;
                    Break;
               end
               else
                   RandomSelectionA := RandomSelectionA - Population[j].FitnessScore;
          end;
          {Select B}
          RandomSelectionB := Round(Random * TotalFitness); // !!!Remove Rounding!!!
          for j := 0 to NumberOfPopulation - 1 do
          begin
               if (RandomSelectionB < Population[j].FitnessScore) then
               begin
                    for k := 0 to Population[j].NumberOfValues - 1 do
                        SelectionB[k] := Population[j].Values[k];
                    BFitness := Population[j].FitnessScore;
                    Break;
               end
               else
                   RandomSelectionB := RandomSelectionB - Population[j].FitnessScore;
          end;

     {- Cross-Over Rate -}
          if (Random <= CrossOver_Rate) then
          begin
               SplitPos := Round(Random * Population[i].NumberOfValues);
               {First Half from A}
               for j := 0 to SplitPos - 1 do
                   NewPopulation[i].Values[j] := SelectionA[j];
               {Second Half from B}
               for j := SplitPos to Population[i].NumberOfValues - 1 do
                   NewPopulation[i].Values[j] := SelectionB[j];
          end
          else
          begin
               if (AFitness > BFitness) then
                  for j := 0 to NewPopulation[i].NumberOfValues - 1 do
                      NewPopulation[i].Values[j] := SelectionA[j]
               else
                   for j := 0 to NewPopulation[i].NumberOfValues - 1 do
                       NewPopulation[i].Values[j] := SelectionB[j];
          end;

     {- Mutation Rate -}
          for j := 0 to NewPopulation[i].NumberOfValues - 1 do
              if (Random <= Mutation_Rate) then
                 NewPopulation[i].Values[j] := (Random * 2) - 1;
     end;
     {- Update Chromosomes -}
     for i := 0 to NumberOfPopulation - 1 do
         for j := 0 to Population[i].NumberOfValues - 1 do
         begin
              Population[i].Values[j] := NewPopulation[i].Values[j];
              Population[i].FitnessScore := 0;
         end;

     {- Destroy Temporary Objects -}
     for i := 0 to NumberOfPopulation - 1 do
         NewPopulation[i].DeInit;
     NewPopulation := nil;
     SelectionA := nil;
     SelectionB := nil;
end;


{-- Chromosome Functions --}
constructor TChromosome.Init(Size: Integer);
begin
     NumberOfValues := Size;
     SetLength(Values, Size);
     FitnessScore := 0;
end;
destructor TChromosome.DeInit;
begin
     Values := Nil;
end;
procedure TChromosome.GetChromo(var Chromo: Array of Real);
var i: Integer;
begin
     if (Length(Chromo) = NumberOfValues) then
        for i := 0 to NumberOfValues - 1 do
            Chromo[i] := Values[i]
     else
     begin
          ShowMessage('ERROR <TChromosome.GetChromo()>: Non-Equal Array Sizes! tried<' +
                      IntToStr(Length(Chromo)) + '> actual<' + IntToStr(NumberOfValues) + '>');
          Halt;
     end;
end;
procedure TChromosome.PutChromo(Chromo: Array of Real);
var i: Integer;
begin
     if (Length(Chromo) = NumberOfValues) then
        for i := 0 to NumberOfValues - 1 do
            Values[i] := Chromo[i]
     else
     begin
          ShowMessage('ERROR <TChromosome.PutChromo>: Non-Equal Array Sizes! tried<' +
                      IntToStr(Length(Chromo)) + '> actual<' + IntToStr(NumberOfValues) + '>');
          Halt;
     end;
end;


{-- Neuron Functions --}
constructor TNeuron.Init(NumOfInputs: Integer; Digital: Boolean);
var i: Integer;
begin
     NumberOfInputs := NumOfInputs;
     DigitalOutput := Digital;

     {Set Array Size Of Weights}
     SetLength(Weight, NumOfInputs + 1);
     {Init Threshold}
     Weight[0] := (Random * 2) - 1;
     {Init Weights}
     for i := 1 to NumOfInputs do
         Weight[i] := (Random * 2) - 1;
end;
function TNeuron.Output(Inputs: Array of Real): Real;
var i: Integer;
    Activation: Real;
begin
     Activation := 0.0000000001;

     {Sum all weights * inputs}
     for i := 1 to NumberOfInputs do
         Activation := Activation + (Inputs[i - 1] * Weight[i]);

     {Test for Threshold}
     if (DigitalOutput) then
     begin
          if (Activation >= Weight[0]) then
             Result := 1
          else
              Result := 0;
     end
     else
     begin
          Activation := Activation + (Weight[0] * -1);
          {Apply Sigmoid}
          Result := Sigmoid(-1 * Activation, 1.0);
     end;
end;


{-- Neuron Layer Functions --}
constructor TNeuronLayer.Init(NumOfNeurons, NumOfInputsPerNeuron: Integer; Digital: Boolean);
var i: Integer;
begin
     NumberOfNeurons := NumOfNeurons;

     {Set Array Size Of Neurons}
     SetLength(Neurons, NumOfNeurons);
     for i := 0 to NumOfNeurons - 1 do
         Neurons[i] := TNeuron.Init(NumOfInputsPerNeuron, Digital);
end;
procedure TNeuronLayer.Execute(Inputs: Array of Real; var Outputs: Array of Real);
var i: Integer;
    Buffer: Array of Real;
begin
     SetLength(Buffer, NumberOfNeurons);
     for i := 0 to NumberOfNeurons - 1 do
         Buffer[i] := Neurons[i].Output(Inputs);
     for i := 0 to NumberOfNeurons - 1 do
         Outputs[i] := Buffer[i];
     Buffer := Nil;
end;


{-- Neural Net Functions --}
constructor TNeuralNet.Init(NumOfInputs: Integer; NeuronSet: Array of Integer; Digital: Boolean);
var i: Integer;
begin
     DigitalOutput := Digital;
     NumberOfInputs := NumOfInputs;
     NumberOfLayers := Length(NeuronSet);
     NumberOfOutputs := NeuronSet[NumberOfLayers - 1];

     SetLength(Layers, NumberOfLayers);

     Layers[0] := TNeuronLayer.Init(NeuronSet[0], NumberOfInputs, Digital);
     if (NumberOfLayers > 1) then
        for i := 1 to NumberOfLayers - 1 do
            Layers[i] := TNeuronLayer.Init(NeuronSet[i], Layers[i - 1].NumberOfNeurons, Digital);
end;
destructor TNeuralNet.Kill;
var i, j: Integer;
begin
     for i := 0 to NumberOfLayers - 1 do
     begin
          for j := 0 to Layers[i].NumberOfNeurons - 1 do
          begin
               Layers[i].Neurons[j].Weight := Nil;
               Layers[i].Neurons[j].Free;
          end;
          Layers[i].Neurons := Nil;
          Layers[i].Free;
     end;
     Layers := Nil;
end;
function TNeuralNet.GetNumberOfWeights: Integer;
var i, j: Integer;
begin
     Result := 0;
     if (NumberOfLayers > 0) then
        for i := 0 to NumberOfLayers - 1 do
            if (Layers[i].NumberOfNeurons > 0) then
               for j := 0 to Layers[i].NumberOfNeurons - 1 do
                   Result := Result + Layers[i].Neurons[j].NumberOfInputs + 1;
end;
procedure TNeuralNet.GetWeights(var Weights: Array of Real);
var i, j, k, w: Integer;
begin
//GetNumberOfWeights
     w := 0;
     if (NumberOfLayers > 0) then
        for i := 0 to NumberOfLayers - 1 do
            if (Layers[i].NumberOfNeurons > 0) then
               for j := 0 to Layers[i].NumberOfNeurons - 1 do
                   for k := 0 to Layers[i].Neurons[j].NumberOfInputs do
                   begin
                        Weights[w] := Layers[i].Neurons[j].Weight[k];
                        inc(w);
                   end;
end;
procedure TNeuralNet.PutWeights(Weights: Array of Real);
var i, j, k, w: Integer;
begin
     w := 0;
     if (NumberOfLayers > 0) then
        for i := 0 to NumberOfLayers - 1 do
            if (Layers[i].NumberOfNeurons > 0) then
               for j := 0 to Layers[i].NumberOfNeurons - 1 do
                   for k := 0 to Layers[i].Neurons[j].NumberOfInputs do
                   begin
                        Layers[i].Neurons[j].Weight[k] := Weights[w];
                        inc(w);
                   end;
end;
procedure TNeuralNet.Execute(Input: Array of Real; var Output: Array of Real);
var i, j: Integer;
    inBuffer, outBuffer: Array of Real;
begin
     if (Length(Input) = NumberOfInputs) then
     begin
          SetLength(outBuffer, Layers[0].NumberOfNeurons);
          Layers[0].Execute(Input, outBuffer);
          if (NumberOfLayers > 1) then
             for i := 1 to NumberOfLayers - 1 do
             begin
                  SetLength(inBuffer, Layers[i - 1].NumberOfNeurons);
                  for j := 0 to Length(inBuffer) - 1 do
                      inBuffer[j] := outBuffer[j];

                  SetLength(outBuffer, Layers[i].NumberOfNeurons);
                  Layers[i].Execute(inBuffer, outBuffer);
             end;
          {Resolve Output}
          for i := 0 to NumberOfOutputs - 1 do
              if (DigitalOutput) then
              begin
                   if (outBuffer[i] > 0) then
                      Output[i] := 1
                   else
                       Output[i] := 0;
              end
              else
                  Output[i] := outBuffer[i];
     end;
     inBuffer := Nil;
     outBuffer := Nil;
end;

end.
