unit AdvancedMathUnit;

interface

uses Windows, Math;

const
  {---Math Pre-Calculations---}
  Pi                 = 3.1415926535897932384626433832795;
  Pi_Div_180         = Pi / 180;       {Multiply by this for Convertion to Radians}
  _180_Div_Pi        = 180 / Pi;       {Multiply by this for Convertion to Degrees}
  Pi_Div_180_Mul_360 = 360 * Pi / 180; {360 Degrees in Radians}

var
  {--- Trig Lookup Tables ---}
  getSIN,
  getCOS,
  getTAN,
  getArcCOS,
  getArcSIN,
  getArcTAN: Array[-360 .. 360] of Extended;


{Pre-Calculate SIN and COS Tables}
procedure PreCalculate_Trig;

{Advanced Trig Functions}
function GetDistance(X1, Y1, X2, Y2: Extended): Extended;

{Angle of Dest Point in Degrees and Radians}
function GetAngleToDestinationDeg(OriginX, OriginY, DestX, DestY: Extended): Integer;
function GetAngleToDestinationRad(OriginX, OriginY, DestX, DestY: Extended): Extended;

{Returns True if Angle is in between LeftArch and RightArch}
function isBetweenArchesDeg(Angle, LeftArch, RightArch: Integer): Boolean;

{XY 2-D Rotation for Degrees and Radians}
function RotateXDeg(oX, oY: Extended; AngleinDegrees: Integer): Extended;
function RotateXRad(oX, oY, AngleinRadians: Extended): Extended;
function RotateYDeg(oX, oY: Extended; AngleinDegrees: Integer): Extended;
function RotateYRad(oX, oY, AngleinRadians: Extended): Extended;
function RotateXYDeg(oX, oY: Extended; AngleinDegrees: Integer): TPoint;
function RotateXYRad(oX, oY, AngleinRadians: Extended): TPoint;

{Convertions}
function DegreesToRadians(Deg: Extended): Extended;
function RadiansToDegrees(Rad: Extended): Extended;
function DegreesTo256(Deg: Integer): Integer;

implementation

{Pre-Calculate SIN and COS Tables}
procedure PreCalculate_Trig;
var i: Integer;
begin
     for i := Low(getSIN) to High(getSIN) do
         getSIN[i] := sin(i * Pi_Div_180);
     for i := Low(getCOS) to High(getCOS) do
         getCOS[i] := cos(i * Pi_Div_180);
     for i := Low(getTAN) to High(getTAN) do
         getTAN[i] := tan(i * Pi_Div_180);

//     for i := Low(getArcSIN) to High(getArcSIN) do
//         getArcSIN[i] := ArcSin(i * Pi_Div_180);
//     for i := Low(getArcCOS) to High(getArcCOS) do
//         getArcCOS[i] := ArcCos(i * Pi_Div_180);
//     for i := Low(getArcTAN) to High(getArcTAN) do
//         getArcTAN[i] := ArcTan(i * Pi_Div_180);
end;


{Get Distance of two Points}
function GetDistance(X1, Y1, X2, Y2: Extended): Extended;
begin
//     Result := sqrt(sqr(X2 - X1) + sqr(Y2 - Y1));
     Result := sqrt((X2 - X1) * (X2 - X1) + (Y2 - Y1) * (Y2 - Y1));
end;


{Angle of Dest Point in Degrees and Radians}
function GetAngleToDestinationDeg(OriginX, OriginY, DestX, DestY: Extended): Integer;
var XDiff: Extended;
begin
     XDiff := DestX - OriginX;
     if (XDiff = 0) then
        XDiff := 0.0001; // XDiff Cannot be Zero!

     if (OriginY > DestY) then
        Result := Round(360 - RadToDeg(ArcCos(XDiff / GetDistance(OriginX, OriginY, DestX, DestY))))
     else
         Result := Round(RadToDeg(ArcCos(XDiff / GetDistance(OriginX, OriginY, DestX, DestY))));
end;
function GetAngleToDestinationRad(OriginX, OriginY, DestX, DestY: Extended): Extended;
var XDiff: Extended;
begin
     XDiff := DestX - OriginX;
     if (XDiff = 0) then
        XDiff := 0.0001; // XDiff Cannot be Zero!

     if (OriginY > DestY) then
        Result := Pi_Div_180_Mul_360 - ArcCos(XDiff / GetDistance(OriginX, OriginY, DestX, DestY))
     else
         Result := ArcCos(XDiff / GetDistance(OriginX, OriginY, DestX, DestY));
end;


{Returns True if Angle is in between LeftArch and RightArch}
function isBetweenArchesDeg(Angle, LeftArch, RightArch: Integer): Boolean;
var CorrectedAngle,
    CorrectedLeftArch,
    CorrectedRightArch: Integer;
begin
     CorrectedAngle := Angle;
     CorrectedLeftArch := LeftArch;
     CorrectedRightArch := RightArch;

     {Make sure Angle is a proper value}
     while (CorrectedAngle > 360) and (CorrectedAngle < 0) do
     begin
          if (CorrectedAngle > 360) then
             dec(CorrectedAngle, 360);
          if (CorrectedAngle < 0) then
             inc(CorrectedAngle, 360);
     end;
     {Make sure LeftArch is a proper value}
     while (CorrectedLeftArch > 360) and (CorrectedLeftArch < 0) do
     begin
          if (CorrectedLeftArch > 360) then
             dec(CorrectedLeftArch, 360);
          if (CorrectedLeftArch < 0) then
             inc(CorrectedLeftArch, 360);
     end;
     {Make sure RightArch is a proper value}
     while (CorrectedRightArch > 360) and (CorrectedRightArch < 0) do
     begin
          if (CorrectedRightArch > 360) then
             dec(CorrectedRightArch, 360);
          if (CorrectedRightArch < 0) then
             inc(CorrectedRightArch, 360);
     end;

     Result := False;
     {Check that LeftArch is lower than RightArch}
     if (CorrectedLeftArch > CorrectedRightArch) then
        if (CorrectedAngle >= CorrectedLeftArch) or
           (CorrectedAngle <= CorrectedRightArch) then
           Result := True
     else
         if (CorrectedAngle >= CorrectedLeftArch) and
            (CorrectedAngle <= CorrectedRightArch) then
            Result := True;
end;


{XY 2-D Rotation for Degrees and Radians}
{X Only}
function RotateXDeg(oX, oY: Extended; AngleinDegrees: Integer): Extended;
begin
     Result := oX * getCOS[AngleinDegrees] - oY * getSIN[AngleinDegrees];
end;
function RotateXRad(oX, oY, AngleinRadians: Extended): Extended;
begin
     Result := oX * cos(AngleinRadians) - oY * sin(AngleinRadians);
end;
{Y Only}
function RotateYDeg(oX, oY: Extended; AngleinDegrees: Integer): Extended;
begin
     Result := oX * getSIN[AngleinDegrees] + oY * getCOS[AngleinDegrees];
end;
function RotateYRad(oX, oY, AngleinRadians: Extended): Extended;
begin
     Result := oX * sin(AngleinRadians) + oY * cos(AngleinRadians);
end;
{Both XY}
function RotateXYDeg(oX, oY: Extended; AngleinDegrees: Integer): TPoint;
begin
     Result.x := Round(oX * getCOS[AngleinDegrees] - oY * getSIN[AngleinDegrees]);
     Result.y := Round(oX * getSIN[AngleinDegrees] + oY * getCOS[AngleinDegrees]);
end;
function RotateXYRad(oX, oY, AngleinRadians: Extended): TPoint;
begin
     Result.x := Round(oX * cos(AngleinRadians) - oY * sin(AngleinRadians));
     Result.y := Round(oX * sin(AngleinRadians) + oY * cos(AngleinRadians));
end;

{Convertions}
{Convert Degrees to Radians}
function DegreesToRadians(Deg: Extended): Extended;
begin
     Result := Deg * Pi_Div_180;
end;
{Convert Radians to Degrees}
function RadiansToDegrees(Rad: Extended): Extended;
begin
     Result := Rad * _180_Div_Pi;
end;
{Convert Degrees to 256 Range}
function DegreesTo256(Deg: Integer): Integer;
begin
     if (Deg = 0) then
        Result := 0
     else
         Result := Round((Deg / 360) * 256);
end;


end.
